"""
This is a class function that drops the pv systems with generates power over night.
"""
import logging

import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("drop_night_pv")
class DropPvSysGeneratingOvernightIterDataPipe(IterDataPipe):
    """
    Drop the pv systems which generates power over night date from a timeseries xarray Dataset.
    """

    def __init__(self, source_datapipe: IterDataPipe):
        """
        This method drops the PV systems producing output overnight

        Args:
            source_datapipe: A datapipe that emmits Xarray Dataset of the pv.netcdf file
        """
        self.source_datapipe = source_datapipe

    def __iter__(self) -> xr.DataArray():

        logger.warning("This droping of the nighttime pv is only applicable to the UK PV datasets")

        for xr_dataset in self.source_datapipe:

            # Getting the list of the pv system id
            pv_sys_id_list = xr_dataset.coords["pv_system_id"].values

            # Grouping the datastet with only night status
            night_ds = xr_dataset.groupby("status_daynight")["night"].values

            # Getting the shape of the array
            night_ds_shape = night_ds.shape

            # Checking if all the systems has any values other than zero (including NaN)
            check_nonzero = [(np.array(night_ds[:, m]) != 0) for m in range(night_ds_shape[1])]

            # Checking if all the systems has NaN in their daily outputs
            check_isnan = [
                np.logical_not(np.isnan(night_ds[:, m])) for m in range(night_ds_shape[1])
            ]

            # Checking if there are any systems which has numeric values (not including np.zeros)
            # and (not including np.nan)
            final_check = np.logical_and(check_nonzero, check_isnan)

            # Getting the indices of system ids to drop
            drop_pv_sys_id = np.where(final_check.any(axis=1))[0]

            logger.info(
                f"Dropping the pv systems{pv_sys_id_list[drop_pv_sys_id]} with night time pv output"
            )
            xr_dataset = xr_dataset.drop_sel(pv_system_id=pv_sys_id_list[drop_pv_sys_id])
            yield xr_dataset



"""Select channels"""
import logging
from typing import List, Union

import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("select_channels")
class SelectChannelsIterDataPipe(IterDataPipe):
    """Select channels"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        channels: List[str],
        dim_name: str = "channel",
    ):
        """
        Select channels

        Args:
            source_datapipe: Datapipe of Xarray objects
            channels: Channel names to keep
            dim_name: Dimension name for time
        """
        self.source_datapipe = source_datapipe
        self.channels = channels
        self.dim_name = dim_name

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for xr_data in self.source_datapipe:
            logger.debug(f"Selecting Channels: {self.channels} out of {xr_data['channel'].values}")
            xr_data = xr_data.sel({self.dim_name: list(self.channels)})
            yield xr_data



"""Downsample Xarray datasets Datapipe"""
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe


@functional_datapipe("downsample")
class DownsampleIterDataPipe(IterDataPipe):
    """Downsample Xarray dataset with coarsen"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        y_coarsen: int,
        x_coarsen: int,
        x_dim_name: str = "x_osgb",
        y_dim_name: str = "y_osgb",
    ):
        """
        Downsample xarray dataset/dataarrays with coarsen

        Args:
            source_datapipe: Datapipe emitting Xarray dataset
            y_coarsen: Coarsen value in the y direction
            x_coarsen: Coarsen value in the x direction
            x_dim_name: X dimension name
            y_dim_name: Y dimension name
        """
        self.source_datapipe = source_datapipe
        self.y_coarsen = y_coarsen
        self.x_coarsen = x_coarsen
        self.x_dim_name = x_dim_name
        self.y_dim_name = y_dim_name

    def __iter__(self):
        """Coarsen the data on the specified dimensions"""
        for xr_data in self.source_datapipe:
            yield xr_data.coarsen(
                {self.y_dim_name: self.y_coarsen, self.x_dim_name: self.x_coarsen},
                boundary="trim",
            ).mean()



"""Offset T0 for training to more realistically match production"""
from typing import Union

import numpy as np
import pandas as pd
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe


@functional_datapipe("offset_t0")
class OffsetT0IterDataPipe(IterDataPipe):
    """Offset T0 for training to more realistically match production"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        max_t0_offset_minutes: Union[float, int],
        min_t0_offset_minutes: Union[float, int] = 0.0,
    ):
        """
        Offset T0 for more realistic training

        Args:
            source_datapipe: Datapipe of t0 times
            max_t0_offset_minutes: Max offset to be applied
            min_t0_offset_minutes: Min offset to be applied
        """
        self.source_datapipe = source_datapipe
        self.max_t0_offset_minutes = max_t0_offset_minutes
        self.min_t0_offset_minutes = min_t0_offset_minutes

    def __iter__(self):
        for t0 in self.source_datapipe:
            offset = pd.Timedelta(
                minutes=np.random.randint(self.min_t0_offset_minutes, self.max_t0_offset_minutes)
            )
            yield t0 - offset



"""Ensure there are N NWP variables by tiling the data"""
import numpy as np
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.consts import BatchKey, NumpyBatch


@functional_datapipe("ensure_n_nwp_variables")
class EnsureNNWPVariables(IterDataPipe):
    """Ensure there are N NWP variables by tiling the data"""

    def __init__(self, source_datapipe: IterDataPipe, num_variables: int):
        """
        Ensure there are N NWP variables by tiling the data

        Args:
            source_datapipe: NumpyBatch emitting NWP datapipe
            num_variables: Number of variables needed
        """
        self.source_datapipe = source_datapipe
        self.num_variables = num_variables

    def __iter__(self) -> NumpyBatch:
        """Ensure there are N NWP variables by tiling the data"""
        for np_batch in self.source_datapipe:
            num_tiles = int(np.ceil(self.num_variables / len(np_batch[BatchKey.nwp_channel_names])))
            np_batch[BatchKey.nwp_channel_names] = np.tile(
                np_batch[BatchKey.nwp_channel_names], num_tiles
            )[: self.num_variables]
            np_batch[BatchKey.nwp] = np.tile(np_batch[BatchKey.nwp], (1, 1, num_tiles, 1, 1))[
                :, :, : self.num_variables, :, :
            ]
            yield np_batch



"""Experimental Datapipes

Collection of experimental datapipes, explicitly not ready for production,
but can be upgraded later, these might not be tested, and are to make quick experiments easier
"""
from .ensure_n_nwp_variables import EnsureNNWPVariables
from .set_system_id_to_one import SetSystemIDsToOneIterDataPipe as SetSystemIDsToOne



"""
This is a class function that slices a contigous datetime range to the 12th hour
"""
import logging

import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("trim_dates_with_insufficent_data")
class TrimDatesWithInsufficentDataIterDataPipe(IterDataPipe):
    """Trim the date values of the Xarray Timeseries data"""

    def __init__(self, source_datapipe: IterDataPipe, minimum_number_data_points: int):
        """Trim the dates to the exact 12th hour

        For the five minute interval, If the time_utc dates are insufficent and
        less than multiple of 289, this method trims that extended dates

        For example:
            <xr.DataArray> time_utc : "2020-01-01T00:00"..."2020-01-02T00:00"..."2020-01-02T06:00"

        This method trims the inusfficent less than one day data at
        the end and provides full set of complete one day interval (5min or 15min or...)

        Args:
            source_datapipe: Xarray emitting timeseries data
            minimum_number_data_points: Minimum number of data intervals in a given day
                5min xarray interval data = 288
                15min xarray interval data = 96
                .........
        """
        self.source_datapipe = source_datapipe
        self.intervals = minimum_number_data_points

    def __iter__(self) -> xr.DataArray():
        # This dropping of insufficent data considers just dates in a given datetime
        for xr_dataset in self.source_datapipe:

            # Getting the 'datetime' values into a single 1D array
            dates_array = xr_dataset.coords["time_utc"].values

            # Checking if the total length of 'datetime' is
            # greater than provided time intervals
            if dates_array.size >= self.intervals:

                # Counting the minute intervals (both non_zero and zero),
                # as every 5min, 15min, or 30 min
                # has intervals such as, for 15 min [0, 15, 30, 45, 0,........]
                total_minute_intervals = xr_dataset.time_utc.dt.minute.values
                count_minute_intervals = np.count_nonzero(
                    total_minute_intervals
                ) + np.count_nonzero(total_minute_intervals == 0)
                # Collecting five minute intervals and counting them
                logger.info(f"Total number of those five minutes are {count_minute_intervals}")

                logger.info(
                    f"Checking if the count is a multiple of given interval {self.intervals}"
                )

                # Checking if the minute intervals are multiples of total intervals in a day
                # For example, datet time values of one day with five minute intervals
                # consists of 288 five-minutes
                check = count_minute_intervals % self.intervals == 0.0

                if not check:
                    # Counting number of intervals needed to be trimmed at the end
                    # The check would be always false, as in a given day,
                    # the last time step would be of the next day
                    trim_dates_position = int(count_minute_intervals % self.intervals)

                    # Number of intervals needed to be trimmed
                    # at the end are trim_dates_position
                    trim_dates = dates_array[-trim_dates_position:]
                    logger.info(f"The trimmed dates are as follows {trim_dates}")

                    # Dropping the dates coordinate variable data and its data in the xarray
                    xr_dataset = xr_dataset.drop_sel(time_utc=trim_dates)

            yield xr_dataset



from datetime import timedelta

from ocf_datapipes.select import SelectLiveT0Time
from ocf_datapipes.transform.xarray import ConvertToNWPTargetTime


def test_add_nwp_target_time(nwp_datapipe):
    t0_datapipe = SelectLiveT0Time(nwp_datapipe, dim_name="init_time_utc")
    nwp_datapipe = ConvertToNWPTargetTime(
        nwp_datapipe,
        t0_datapipe,
        sample_period_duration=timedelta(minutes=60),
        history_duration=timedelta(hours=2),
        forecast_duration=timedelta(hours=3),
    )
    data = next(iter(nwp_datapipe))
    assert "target_time_utc" in data.coords
    assert len(data.coords["target_time_utc"]) == 6



"""Add fake future data for proper Power Perceiver Production"""
import numpy as np
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.consts import BatchKey, NumpyBatch


@functional_datapipe("add_zeroed_future_data")
class AddZeroedFutureDataIterDataPipe(IterDataPipe):
    """Extends timestamps into the future"""

    def __init__(self, source_datapipe: IterDataPipe, key: BatchKey, time_key: BatchKey):
        """
        Extends data into the future, zeroed out

        This assumes that the current time_utc array only covers history + now,
        so just extends it further into the future

        Args:
            source_datapipe: Datapipe of NumpyBatch
            key: BatchKey to extend
            time_key: Batch key to use as the time for how many to append
        """
        self.source_datapipe = source_datapipe
        self.key = key
        self.time_key = time_key

    def __iter__(self) -> NumpyBatch:
        for np_batch in self.source_datapipe:
            data = np_batch[self.key]
            new_data = np.zeros((len(np_batch[self.time_key]), *data.shape[1:]), dtype=data.dtype)
            for t in range(data.shape[0]):
                new_data[t] = data[t]
            np_batch[self.key] = new_data
            yield np_batch



"""
Drop PV systems which has only NaN's in a single day
"""
import logging

import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.utils import return_system_indices_which_has_contiguous_nan

logger = logging.getLogger(__name__)


@functional_datapipe("drop_pv_systems_with_only_nan_in_a_day")
class DropPVSystemsWithOnlyNanInADayIterDataPipe(IterDataPipe):
    """Remove any PV systems with less than 1 day of data"""

    def __init__(self, source_datapipe: IterDataPipe, minimum_number_data_points: int) -> None:
        """Remove any PV systems with less than 1 day of data

        This is done, by counting all the non values and check the
        count is greater than 289 (number of 5 minute intervals in a day)

        Args:
            source_datapipe: Datapipe of Xarray Dataset emitting timeseries data
            minimum_number_data_points: Minimum number of intervals in a given day
                For 5min time series, intervals = 288
                For 30min time series, intervals = 48
        """

        self.source_datapipe = source_datapipe
        self.intervals = minimum_number_data_points

    def __iter__(self) -> xr.Dataset():

        # Reading the DataArray
        for xr_dataset in self.source_datapipe:

            dates_array = xr_dataset.coords["time_utc"].values

            # Checking length of time series longer than standard intervals
            if dates_array.size >= self.intervals:

                # Collecting the time series data from 'time_utc' coordinate
                for i in range(0, dates_array.size, self.intervals):

                    if i == self.intervals:
                        break
                    else:

                        # Getting the first and last timestamp of the day based on intervals
                        # Slicing the dataset by individual first timestamp in a day
                        # and last timestamp of the day
                        xr_ds = xr_dataset.sel(
                            time_utc=slice(dates_array[i], dates_array[i + (self.intervals - 1)])
                        )

                        # Extracting system ids with continous NaN's
                        sys_ids = return_system_indices_which_has_contiguous_nan(
                            xr_ds.values, check_interval=self.intervals
                        )

                        # Dropping the systems which are inactive for the whole day
                        if not len(sys_ids) == 0:
                            xr_dataset = xr_dataset.drop_isel(pv_system_id=sys_ids)
            else:
                pass

            yield xr_dataset



import numpy as np

from ocf_datapipes.transform.xarray import Normalize
from ocf_datapipes.utils.consts import NWP_MEAN, NWP_STD, SAT_MEAN, SAT_STD


def test_normalize_sat(sat_datapipe):
    sat_datapipe = Normalize(sat_datapipe, mean=SAT_MEAN["HRV"], std=SAT_STD["HRV"])
    data = next(iter(sat_datapipe))
    assert np.all(data <= 1.0)
    assert np.all(data >= -1.0)


def test_normalize_nwp(nwp_datapipe):
    nwp_datapipe = Normalize(nwp_datapipe, mean=NWP_MEAN, std=NWP_STD)
    data = next(iter(nwp_datapipe))
    assert np.all(data.values <= 1.0)
    # TODO Check why this normalization is so negative
    assert np.all(data.values >= -60.0)


def test_normalize_topo(topo_datapipe):
    topo_datapipe = topo_datapipe.reproject_topography().normalize(
        calculate_mean_std_from_example=True
    )
    data = next(iter(topo_datapipe))
    assert np.all(data >= -1.0)
    # TODO Check why this normalization is ends up with a max of 11ish instead
    assert np.all(data <= 11.0)


def test_normalize_gsp(gsp_datapipe):
    gsp_datapipe = gsp_datapipe.normalize(normalize_fn=lambda x: x / x.capacity_megawatt_power)
    data = next(iter(gsp_datapipe))
    assert np.min(data) >= 0.0
    assert np.max(data) <= 1.0


def test_normalize_passiv(passiv_datapipe):
    passiv_datapipe = passiv_datapipe.normalize(normalize_fn=lambda x: x / x.capacity_watt_power)
    data = next(iter(passiv_datapipe))
    assert np.min(data) >= 0.0
    assert np.max(data) <= 1.0


def test_normalize_pvoutput(pvoutput_datapipe):
    pvoutput_datapipe = pvoutput_datapipe.normalize(
        normalize_fn=lambda x: x / x.capacity_watt_power
    )
    data = next(iter(pvoutput_datapipe))
    assert np.min(data) >= 0.0
    assert np.max(data) <= 1.0



"""
Pipeline to return live GSP and NWP data
"""

import logging
from pathlib import Path
from typing import Union

import xarray

import ocf_datapipes  # noqa
from ocf_datapipes.config.load import load_yaml_configuration
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.load import OpenGSPFromDatabase, OpenNWP
from ocf_datapipes.load.nwp.nwp import OpenLatestNWPDataPipe

logger = logging.getLogger(__name__)
xarray.set_options(keep_attrs=True)


def xgnational_production(configuration_filename: Union[Path, str]) -> dict:
    """
    Create the National XG Boost  using a configuration

    Args:
        configuration_filename: Name of the configuration

    Returns:
        dictionary of 'nwp' and 'gsp' containing xarray for both
    """

    configuration: Configuration = load_yaml_configuration(filename=configuration_filename)

    logger.debug("Opening Datasets")
    base_nwp_datapipe = OpenNWP(configuration.input_data.nwp.nwp_zarr_path)
    nwp_datapipe = OpenLatestNWPDataPipe(base_nwp_datapipe)
    gsp_datapipe = OpenGSPFromDatabase(
        history_minutes=configuration.input_data.gsp.history_minutes,
        interpolate_minutes=configuration.input_data.gsp.live_interpolate_minutes,
        load_extra_minutes=configuration.input_data.gsp.live_load_extra_minutes,
        national_only=True,
    )

    nwp_xr = next(iter(nwp_datapipe))
    gsp_xr = next(iter(gsp_datapipe))

    return {"nwp": nwp_xr, "gsp": gsp_xr}
